import os

from typing import AsyncGenerator, Callable
from dataclasses import dataclass
from io import BytesIO
from http import HTTPMethod

from .request import RequestClient
from .itags import get_format_profile
from .helpers import safe_filename, target_directory
from .extractors import mime_type_codec

DEFAULT_RANGE_SIZE = 9437184


@dataclass
class Stream:
    title: str
    author: str
    request_client: RequestClient
    url: str
    itag: int
    mimeType: str
    bitrate: int
    lastModified: int
    quality: str
    projectionType: str
    approxDurationMs: str
    is_otf: bool
    highReplication: bool | None = None
    loudnessDb: float | None = None
    initRange: dict | None = None
    indexRange: dict | None = None
    colorInfo: dict | None = None
    width: int | None = None
    height: int | None = None
    contentLength: str | None = None
    fps: int | None = None
    qualityLabel: str | None = None
    averageBitrate: int | None = None
    audioQuality: str | None = None
    audioSampleRate: str | None = None
    audioChannels: int | None = None

    def __init__(self, **kwargs):  # for unexcepted
        self.__dict__.update(kwargs)

    @property
    def itag_profile(self):
        return get_format_profile(self.itag)

    @property
    def abr(self):
        return self.itag_profile["abr"]

    @property
    def resolution(self):
        return self.itag_profile["resolution"]

    @property
    def mime_type(self):
        mime_types, codecs = mime_type_codec(self.mimeType)
        return mime_types

    @property
    def codecs(self):
        mime_type, codecs = mime_type_codec(self.mimeType)
        return codecs

    @property
    def type(self):
        return self.mimeType.split("/")[0]

    @property
    def subtype(self):
        return self.mime_type.split("/")[1]

    @property
    def is_adaptive(self) -> bool:
        return bool(len(self.codecs) % 2)

    @property
    def is_progressive(self) -> bool:
        return not self.is_adaptive

    @property
    def includes_audio_track(self) -> bool:
        return self.is_progressive or self.type == "audio"

    @property
    def includes_video_track(self) -> bool:
        return self.is_progressive or self.type == "video"

    async def filesize(self) -> int:
        response = await self.request_client.request(method=HTTPMethod.HEAD, url=self.url)
        return int(response["headers"].get("Content-Length"))

    async def _download(self) -> AsyncGenerator[bytes, None]:
        file_size = await self.filesize()
        downloaded = 0
        while downloaded < file_size:
            stop_pos = min(downloaded + DEFAULT_RANGE_SIZE, file_size) - 1
            range_header = f"bytes={downloaded}-{stop_pos}"
            request = await self.request_client.request(
                method=HTTPMethod.GET, url=self.url, headers={"Range": range_header}
            )
            chunk = request["response"]
            downloaded += len(chunk)
            yield chunk

    async def download_buffer(self) -> BytesIO:
        io = BytesIO()
        async for chunk in self._download():
            io.write(chunk)
        io.seek(0)
        return io

    async def download_filepath(self, filename: str = None, output_path: str = None):
        file_path = self.get_file_path(
            filename=filename,
            output_path=output_path,
        )
        with open(file_path, "wb") as file:
            async for chunk in self._download():
                file.write(chunk)
        return file_path

    @property
    def default_filename(self) -> str:
        filename = safe_filename(self.title)
        return f"{filename}.{self.subtype}"

    def get_file_path(
        self,
        filename: str | None = None,
        output_path: str | None = None,
    ) -> str:
        if not filename:
            filename = self.default_filename
        return os.path.join(target_directory(output_path), filename)


class StreamQuery:
    def __init__(self, fmt_streams: list[Stream]):
        self.fmt_streams = fmt_streams

    def filter(
        self,
        fps=None,
        res=None,
        resolution=None,
        mime_type=None,
        type=None,
        subtype=None,
        file_extension=None,
        abr=None,
        bitrate=None,
        video_codec=None,
        audio_codec=None,
        only_audio=None,
        only_video=None,
        progressive=None,
        adaptive=None,
        is_dash=None,
        custom_filter_functions=None,
    ):
        filters = []
        if res or resolution:
            filters.append(lambda s: s.resolution == (res or resolution))

        if fps:
            filters.append(lambda s: s.fps == fps)

        if mime_type:
            filters.append(lambda s: s.mime_type == mime_type)

        if type:
            filters.append(lambda s: s.type == type)

        if subtype or file_extension:
            filters.append(lambda s: s.subtype == (subtype or file_extension))

        if abr or bitrate:
            filters.append(lambda s: s.abr == (abr or bitrate))

        if video_codec:
            filters.append(lambda s: s.video_codec == video_codec)

        if audio_codec:
            filters.append(lambda s: s.audio_codec == audio_codec)

        if only_audio:
            filters.append(
                lambda s: (s.includes_audio_track and not s.includes_video_track),
            )

        if only_video:
            filters.append(
                lambda s: (s.includes_video_track and not s.includes_audio_track),
            )

        if progressive:
            filters.append(lambda s: s.is_progressive)

        if adaptive:
            filters.append(lambda s: s.is_adaptive)

        if custom_filter_functions:
            filters.extend(custom_filter_functions)

        if is_dash is not None:
            filters.append(lambda s: s.is_dash == is_dash)

        return self._filter(filters)

    def _filter(self, filters: list[Callable]) -> "StreamQuery":
        fmt_streams = self.fmt_streams
        for filter_lambda in filters:
            fmt_streams = filter(filter_lambda, fmt_streams)
        return StreamQuery(list(fmt_streams))

    def order_by(self, attribute_name: str) -> "StreamQuery":
        has_attribute = [
            s for s in self.fmt_streams if getattr(s, attribute_name) is not None
        ]
        if has_attribute and isinstance(getattr(has_attribute[0], attribute_name), str):
            try:
                return StreamQuery(
                    sorted(
                        has_attribute,
                        key=lambda s: int(
                            "".join(filter(str.isdigit, getattr(s, attribute_name)))
                        ),  # type: ignore  # noqa: E501
                    )
                )
            except ValueError:
                pass
        return StreamQuery(
            sorted(has_attribute, key=lambda s: getattr(s, attribute_name))
        )

    def last(self):
        try:
            return self.fmt_streams[-1]
        except IndexError:
            pass

    def get_audio_only(self, subtype: str = "mp4") -> Stream | None:
        return self.filter(only_audio=True, subtype=subtype).order_by("abr").last()

    def get_highest_resolution(self) -> Stream | None:
        return self.filter(progressive=True).order_by("resolution").last()



from http import HTTPMethod

from .request import RequestClient
from .extractors import extract_video_id, playability_status, apply_descrambler
from .exceptions import VideoPrivate, VideoUnavailable, LiveStreamError, MembersOnly, RecordingUnavailable
from .streams import Stream, StreamQuery


class Video:
    def __init__(self, video_url: str):
        self.request_client = RequestClient()
        self.video_id = extract_video_id(video_url)
        self.base_api_url = "https://www.youtube.com/youtubei/v1"
        self.watch_url = f"https://youtube.com/watch?v={self.video_id}"
        self._html: str | None = None
        self._video_data: dict | None = None
        self._fmt_streams: list | None = None

    async def html(self):
        if self._html is None:
            self._html = (await self.request_client.request(
                method=HTTPMethod.GET,
                url=self.watch_url
            ))["response"].decode()
        return self._html

    async def _player(self, request_client: RequestClient):
        endpoint = f"{self.base_api_url}/player"
        query = {"videoId": self.video_id}
        data = query
        return (await request_client.request(
            method=HTTPMethod.POST, url=endpoint, params=query, data=data
        ))["response"]

    async def video_info(self):
        await self.check_availability()
        if self._video_data is None:
            self._video_data = await self._player(self.request_client)
        return self._video_data

    async def streaming_data(self):
        data = await self.video_info()
        if streaming_data := data.get("streamingData"):
            return streaming_data
        else:
            bypassed_data = await self.bypass_age_gate()
            return bypassed_data["streamingData"]

    async def fmt_streams(self):
        await self.check_availability()
        if self._fmt_streams:
            return self._fmt_streams
        self._fmt_streams = []
        stream_manifest = apply_descrambler(await self.streaming_data())
        video_title = await self.title()
        video_author = await self.author()
        for stream in stream_manifest:
            self._fmt_streams.append(
                Stream(title=video_title, author=video_author, request_client=self.request_client, **stream)
            )
        return self._fmt_streams

    async def streams(self) -> StreamQuery:
        return StreamQuery(await self.fmt_streams())

    async def bypass_age_gate(self) -> dict:
        for r_client in ("android_embedded", "ios"):
            request_client = RequestClient(r_client)
            video_data = await self._player(request_client)
            if video_data.get("streamingData"):
                self._video_data = video_data
                return self._video_data

    async def check_availability(self):
        html = await self.html()
        status, messages = playability_status(html)
        # print(messages)
        for reason in messages:
            if status == "UNPLAYABLE":
                if reason == (
                    "Join this channel to get access to members-only content "
                    "like this video, and other exclusive perks."
                ):
                    raise MembersOnly(video_id=self.video_id)
                elif reason == "This live stream recording is not available.":
                    raise RecordingUnavailable(video_id=self.video_id)
                else:
                    raise VideoUnavailable(video_id=self.video_id)
            elif status == "LOGIN_REQUIRED":
                if reason == (
                    "This is a private video. "
                    "Please sign in to verify that you may see it."
                ):
                    raise VideoPrivate(video_id=self.video_id)
            elif status == "ERROR":
                if reason == "Video unavailable":
                    raise VideoUnavailable(video_id=self.video_id)
            elif status == "LIVE_STREAM":
                raise LiveStreamError(video_id=self.video_id)

    async def title(self) -> str:
        """Get the video title."""
        await self.check_availability()
        return (await self.video_info()).get("videoDetails", {}).get("title", "untitled")

    async def author(self) -> str:
        """Get the video author."""
        return (
            (await self.video_info()).get("videoDetails", {}).get("author", "unknown")
        )

    async def length(self) -> int:
        """Get the video length in seconds."""
        return int(
            (await self.video_info()).get("videoDetails", {}).get("lengthSeconds", 0)
        )

    async def thumbnail(self) -> str:
        thumbnails = (await self.video_info()).get("videoDetails", {}).get("thumbnail", {}).get("thumbnails")
        if thumbnails:
            return thumbnails[-1]["url"]
        return f"https://img.youtube.com/vi/{self.video_id}/maxresdefault.jpg"



import json

from http import HTTPMethod
from urllib.parse import urlencode
from aiohttp import ClientSession
from aiohttp.client_exceptions import ContentTypeError

INNERTUBE_CLIENTS = {
    'web': {
        'INNERTUBE_API_KEY': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'WEB',
                'clientVersion': '2.20220801.00.00',
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 1
    },
    'web_embedded': {
        'INNERTUBE_API_KEY': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'WEB_EMBEDDED_PLAYER',
                'clientVersion': '1.20220731.00.00',
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 56
    },
    'web_music': {
        'INNERTUBE_API_KEY': 'AIzaSyC9XL3ZjWddXya6X74dJoCTL-WEYFDNX30',
        'INNERTUBE_HOST': 'music.youtube.com',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'WEB_REMIX',
                'clientVersion': '1.20220727.01.00',
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 67,
    },
    'web_creator': {
        'INNERTUBE_API_KEY': 'AIzaSyBUPetSUmoZL-OhlxA7wSac5XinrygCqMo',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'WEB_CREATOR',
                'clientVersion': '1.20220726.00.00',
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 62,
    },
    'android': {
        'INNERTUBE_API_KEY': 'AIzaSyA8eiZmM1FaDVjRy-df2KTyQ_vz_yYM39w',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'ANDROID',
                'clientVersion': '17.31.35',
                'androidSdkVersion': 30,
                'userAgent': 'com.google.android.youtube/17.31.35 (Linux; U; Android 11) gzip'
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 3,
        'REQUIRE_JS_PLAYER': False
    },
    'android_embedded': {
        'INNERTUBE_API_KEY': 'AIzaSyCjc_pVEDi4qsv5MtC2dMXzpIaDoRFLsxw',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'ANDROID_EMBEDDED_PLAYER',
                'clientVersion': '17.31.35',
                'androidSdkVersion': 30,
                'userAgent': 'com.google.android.youtube/17.31.35 (Linux; U; Android 11) gzip'
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 55,
        'REQUIRE_JS_PLAYER': False
    },
    'android_music': {
        'INNERTUBE_API_KEY': 'AIzaSyAOghZGza2MQSZkY_zfZ370N-PUdXEo8AI',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'ANDROID_MUSIC',
                'clientVersion': '5.16.51',
                'androidSdkVersion': 30,
                'userAgent': 'com.google.android.apps.youtube.music/5.16.51 (Linux; U; Android 11) gzip'
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 21,
        'REQUIRE_JS_PLAYER': False
    },
    'android_creator': {
        'INNERTUBE_API_KEY': 'AIzaSyD_qjV8zaaUMehtLkrKFgVeSX_Iqbtyws8',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'ANDROID_CREATOR',
                'clientVersion': '22.30.100',
                'androidSdkVersion': 30,
                'userAgent': 'com.google.android.apps.youtube.creator/22.30.100 (Linux; U; Android 11) gzip'
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 14,
        'REQUIRE_JS_PLAYER': False
    },
    'ios': {
        'INNERTUBE_API_KEY': 'AIzaSyB-63vPrdThhKuerbB2N_l7Kwwcxj6yUAc',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'IOS',
                'clientVersion': '17.33.2',
                'deviceModel': 'iPhone14,3',
                'userAgent': 'com.google.ios.youtube/17.33.2 (iPhone14,3; U; CPU iOS 15_6 like Mac OS X)'
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 5,
        'REQUIRE_JS_PLAYER': False
    },
    'ios_embedded': {
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'IOS_MESSAGES_EXTENSION',
                'clientVersion': '17.33.2',
                'deviceModel': 'iPhone14,3',
                'userAgent': 'com.google.ios.youtube/17.33.2 (iPhone14,3; U; CPU iOS 15_6 like Mac OS X)'
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 66,
        'REQUIRE_JS_PLAYER': False
    },
    'ios_music': {
        'INNERTUBE_API_KEY': 'AIzaSyBAETezhkwP0ZWA02RsqT1zu78Fpt0bC_s',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'IOS_MUSIC',
                'clientVersion': '5.21',
                'deviceModel': 'iPhone14,3',
                'userAgent': 'com.google.ios.youtubemusic/5.21 (iPhone14,3; U; CPU iOS 15_6 like Mac OS X)'
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 26,
        'REQUIRE_JS_PLAYER': False
    },
    'ios_creator': {
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'IOS_CREATOR',
                'clientVersion': '22.33.101',
                'deviceModel': 'iPhone14,3',
                'userAgent': 'com.google.ios.ytcreator/22.33.101 (iPhone14,3; U; CPU iOS 15_6 like Mac OS X)'
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 15,
        'REQUIRE_JS_PLAYER': False
    },
    'mweb': {
        'INNERTUBE_API_KEY': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'MWEB',
                'clientVersion': '2.20220801.00.00',
            }
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 2
    },
    'tv_embedded': {
        'INNERTUBE_API_KEY': 'AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8',
        'INNERTUBE_CONTEXT': {
            'client': {
                'clientName': 'TVHTML5_SIMPLY_EMBEDDED_PLAYER',
                'clientVersion': '2.0',
            },
        },
        'INNERTUBE_CONTEXT_CLIENT_NAME': 85
    },
}


class RequestClient:
    def __init__(self, client_name: str = "android_music"):
        self.client = INNERTUBE_CLIENTS[client_name]
        self.context = self.client["INNERTUBE_CONTEXT"]
        self.api_key = self.client["INNERTUBE_API_KEY"]

    def base_data(self) -> dict:
        data = {
            "context": self.context,
            "params": "8AEB",
            "playbackContext": {
                "contentPlaybackContext":
                    {"html5Preference": "HTML5_PREF_WANTS"}
            },
            "contentCheckOk": True,
            "racyCheckOk": True
        }
        return data

    def base_headers(self) -> dict:
        headers = {
            "User-Agent": self.context["client"].get("userAgent"),
            "accept-language": "en-US,en",
            "content-type": "application/json",
            "X-YouTube-Client-Name": "3",
            "X-YouTube-Client-Version": self.context["client"]["clientVersion"],
            "Origin": "https://www.youtube.com"
        }
        return headers

    async def request(
        self,
        method: HTTPMethod,
        url: str,
        headers: dict | None = None,
        params: dict | None = None,
        data: dict | None = None
    ):
        base_headers = self.base_headers()
        if method == HTTPMethod.POST:
            base_data = self.base_data()
            if data:
                base_data.update(data)
            base_data = json.dumps(base_data).encode("utf-8")
        else:
            base_data = None
        if params:
            params = urlencode(params)
        if headers:
            base_headers.update(headers)
        async with ClientSession() as session:
            async with session.request(
                method=method,
                url=url,
                headers=headers,
                data=base_data,
                params=params
            ) as resp:
                headers = resp.headers
                try:
                    response = await resp.json()
                except ContentTypeError:
                    response = await resp.read()
        return {"response": response, "headers": headers}
